from contextlib import contextmanager
import hashlib
import os
from pathlib import Path
from tarfile import TarFile
from tempfile import TemporaryDirectory


class _DataFactory:
    """Utility function to generate a unique data string using an incrementing counter."""

    def __init__(self) -> None:
        self.counter = 0

    def __call__(self) -> str:
        out = f"{self.counter:04d}"
        self.counter += 1
        return out


data_factory = _DataFactory()


def file_factory(parent_dir: Path) -> Path:
    """Utility function to generate a file with random data."""
    path = parent_dir / f"{data_factory()}.txt"
    path.write_text(data_factory())
    return path


def dir_tree_factory(parent_dir: Path):
    """Utility function to generate a directory tree containing files with random data."""
    root_dir = parent_dir / data_factory()
    root_dir.mkdir()

    for _ in range(3):
        file_factory(root_dir)

    sub_dir = root_dir / "sub_dir"
    sub_dir.mkdir()

    for _ in range(3):
        file_factory(sub_dir)

    return root_dir


@contextmanager
def umask(mask: int):
    """Utility context manager to temporarily set umask to a new value."""
    old_mask = os.umask(mask)
    yield mask
    os.umask(old_mask)


def hash_file(path: Path):
    """Utility function to calculate the hash of a file's contents."""
    return hashlib.md5(path.read_bytes()).hexdigest()


def assert_archive_contents_equals(arc1: Path, arc2: Path):
    with TemporaryDirectory() as outdir1, TemporaryDirectory() as outdir2:
        try:
            with TarFile.open(arc1, "r") as tp:
                tp.extractall(outdir1, filter="tar")
            with TarFile.open(arc2, "r") as tp:
                tp.extractall(outdir2, filter="tar")
        except TypeError:
            with TarFile.open(arc1, "r") as tp:
                tp.extractall(outdir1)
            with TarFile.open(arc2, "r") as tp:
                tp.extractall(outdir2)

        extracted1 = sorted(Path(outdir1).glob("**/*"))
        extracted2 = sorted(Path(outdir2).glob("**/*"))

        assert [p.relative_to(outdir1) for p in extracted1] == [
            p.relative_to(outdir2) for p in extracted2
        ]
        for arc1_member, arc2_member in zip(extracted1, extracted2):
            assert arc1_member.is_file() == arc2_member.is_file()
            if arc1_member.is_file():
                assert hash_file(arc1_member) == hash_file(arc2_member)



"""A demo script that shows repro_tarfile's reproducible output in contrast with the standard
library module tarfile's nonreproducible output.
"""

import hashlib
from pathlib import Path
import tarfile
from tempfile import TemporaryDirectory
from time import sleep

import repro_tarfile

example_dir = Path(__file__).parent


def create_archive(tar_module, outfile: Path):
    with tar_module.open(outfile, "w:gz") as tar:
        # Use write to add a file to the archive
        tar.add(example_dir / "data.txt", arcname="data.txt")
        # Or writestr to write data to the archive
        tar.addfile(tarfile.TarInfo(name="lore.txt"), fileobj="goodbye")


with TemporaryDirectory() as tempdir_name:
    tempdir = Path(tempdir_name)

    cases = [
        (tarfile, tempdir / "tarfile-1.tar.gz"),
        (tarfile, tempdir / "tarfile-2.tar.gz"),
        (repro_tarfile, tempdir / "repro_tarfile-1.tar.gz"),
        (repro_tarfile, tempdir / "repro_tarfile-2.tar.gz"),
    ]

    for case in cases:
        create_archive(*case)
        sleep(2)

    for _, outfile in cases:
        print(outfile.name, hashlib.md5(outfile.read_bytes()).hexdigest())



import os
from pathlib import Path
import sys

import pytest
from pytest_cases import fixture_union


@pytest.fixture
def abs_path(tmp_path):
    """Fixture that returns a temporary directory as an absolute Path object."""
    return tmp_path


@pytest.fixture
def rel_path(tmp_path):
    """Fixture that sets a temporary directory as the current working directory and returns a
    relative path to it."""
    orig_wd = Path.cwd()
    os.chdir(tmp_path)
    yield Path()
    os.chdir(orig_wd)


# Minimum versions with extractall filter support. Don't test abs_path if we don't have it.
EXTRACTALL_FILTER_MIN_VERSIONS = {
    (3, 8): (3, 8, 17),
    (3, 9): (3, 9, 17),
    (3, 10): (3, 10, 12),
    (3, 11): (3, 11, 4),
    (3, 12): (3, 12),
}
if sys.version_info >= EXTRACTALL_FILTER_MIN_VERSIONS[sys.version_info[:2]]:
    base_path = fixture_union("base_path", ["rel_path", "abs_path"])
else:
    base_path = fixture_union("base_path", ["rel_path"])



from io import StringIO
import platform
import sys
from tarfile import TarFile, TarInfo
from time import sleep

try:
    from time import tzset
except ImportError:
    tzset = None

import pytest

from repro_tarfile import ReproducibleTarFile, mtime
from tests.utils import (
    assert_archive_contents_equals,
    data_factory,
    dir_tree_factory,
    file_factory,
    hash_file,
    umask,
)


def test_add_dir_tree_mtime(base_path):
    """Archiving a directory tree works with different modified time."""
    dir_tree = dir_tree_factory(base_path)

    # Create base ReproducibleTarFile archive
    rptf_arc1 = base_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # Create regular TarFile archive for comparison
    tf_arc1 = base_path / "tf_arc1.tar"
    with TarFile.open(tf_arc1, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # Sleep to update modified times, change permissions
    sleep(2)
    for path in dir_tree.glob("**/*"):
        path.touch()

    # Create second ReproducibleTarFile archive after delay
    rptf_arc2 = base_path / "tptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # Create second regular TarFile archive for comparison after delay
    tf_arc2 = base_path / "tf_arc2.tar"
    with TarFile.open(tf_arc2, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_dir_tree_mode(base_path):
    """Archiving a directory tree works with different permission modes."""
    with umask(0o022):
        dir_tree = dir_tree_factory(base_path)

    # Create base ReproducibleTarFile archive
    rptf_arc1 = base_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # Create regular TarFile archive for comparison
    tf_arc1 = base_path / "zipfile1.tar"
    with TarFile.open(tf_arc1, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # Change permissions
    with umask(0o002) as mask:
        dir_tree.chmod(mode=0o777 ^ mask)
        for path in dir_tree.glob("**/*"):
            if path.is_file():
                path.chmod(mode=0o666 ^ mask)
            else:
                path.chmod(mode=0o777 ^ mask)

    # Create second ReproducibleTarFile archive using files with new permissions
    rptf_arc2 = base_path / "rptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # Create second regular TarFile archive for comparison using files with new permissions
    tf_arc2 = base_path / "zipfile2.tar"
    with TarFile.open(tf_arc2, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(path)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    if platform.system() != "Windows":
        # Windows doesn't seem to actually make them different
        assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_dir_tree_string_paths(rel_path):
    """Archiving a directory tree with string path names."""
    dir_tree = dir_tree_factory(rel_path)

    # Create base ReproducibleTarFile archive
    rptf_arc1 = rel_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(str(path))

    # Create regular TarFile archive for comparison
    tf_arc1 = rel_path / "zipfile1.tar"
    with TarFile.open(tf_arc1, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(str(path))

    # Update modified times
    sleep(2)
    for path in dir_tree.glob("**/*"):
        path.touch()

    # Create second ReproducibleTarFile archive after delay
    rptf_arc2 = rel_path / "rptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(str(path))

    # Create second regular TarFile archive for comparison after delay
    tf_arc2 = rel_path / "zipfile2.tar"
    with TarFile.open(tf_arc2, "w") as tp:
        for path in sorted(dir_tree.glob("**/*")):
            tp.add(str(path))

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_single_file(base_path):
    """Writing the same file with different mtime produces the same hash."""
    data_file = file_factory(base_path)

    rptf_arc1 = base_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        tp.add(data_file)

    tf_arc1 = base_path / "tf_arc1.tar"
    with TarFile.open(tf_arc1, "w") as tp:
        tp.add(data_file)

    print(data_file.stat())
    sleep(2)
    data_file.touch()
    print(data_file.stat())

    rptf_arc2 = base_path / "rptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        tp.add(data_file)

    tf_arc2 = base_path / "tf_arc2.tar"
    with TarFile.open(tf_arc2, "w") as tp:
        tp.add(data_file)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


@pytest.mark.skipif(tzset is None, reason="tzset not available")
def test_mtime_not_affected_by_timezone(monkeypatch):
    monkeypatch.setenv("TZ", "America/Chicago")
    tzset()
    dt1 = mtime()
    monkeypatch.setenv("TZ", "America/Los_Angeles")
    tzset()
    dt2 = mtime()

    assert dt1 == dt2


def test_add_single_file_gz(base_path):
    """Writing the same file with different mtime produces the same hash with gzip compression."""
    data_file = file_factory(base_path)

    rptf_arc1 = base_path / "rptf_arc1.tar.gz"
    with ReproducibleTarFile.open(rptf_arc1, "w:gz") as tp:
        tp.add(data_file)

    tf_arc1 = base_path / "tf_arc1.tar.gz"
    with TarFile.open(tf_arc1, "w:gz") as tp:
        tp.add(data_file)

    print(data_file.stat())
    sleep(2)
    data_file.touch()
    print(data_file.stat())

    rptf_arc2 = base_path / "rptf_arc2.tar.gz"
    with ReproducibleTarFile.open(rptf_arc2, "w:gz") as tp:
        tp.add(data_file)

    tf_arc2 = base_path / "tf_arc2.tar.gz"
    with TarFile.open(tf_arc2, "w:gz") as tp:
        tp.add(data_file)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_single_file_bz2(base_path):
    """Writing the same file with different mtime produces the same hash with bzip2 compression."""
    data_file = file_factory(base_path)

    rptf_arc1 = base_path / "rptf_arc1.tar.bz2"
    with ReproducibleTarFile.open(rptf_arc1, "w:bz2") as tp:
        tp.add(data_file)

    tf_arc1 = base_path / "tf_arc1.tar.bz2"
    with TarFile.open(tf_arc1, "w:bz2") as tp:
        tp.add(data_file)

    print(data_file.stat())
    sleep(2)
    data_file.touch()
    print(data_file.stat())

    rptf_arc2 = base_path / "rptf_arc2.tar.bz2"
    with ReproducibleTarFile.open(rptf_arc2, "w:bz2") as tp:
        tp.add(data_file)

    tf_arc2 = base_path / "tf_arc2.tar.bz2"
    with TarFile.open(tf_arc2, "w:bz2") as tp:
        tp.add(data_file)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_single_file_xz(base_path):
    """Writing the same file with different mtime produces the same hash with xz format /
    LCMA2 compression.
    """
    data_file = file_factory(base_path)

    rptf_arc1 = base_path / "rptf_arc1.tar.xz"
    with ReproducibleTarFile.open(rptf_arc1, "w:xz") as tp:
        tp.add(data_file)

    tf_arc1 = base_path / "tf_arc1.tar.xz"
    with TarFile.open(tf_arc1, "w:xz") as tp:
        tp.add(data_file)

    print(data_file.stat())
    sleep(2)
    data_file.touch()
    print(data_file.stat())

    rptf_arc2 = base_path / "rptf_arc2.tar.xz"
    with ReproducibleTarFile.open(rptf_arc2, "w:xz") as tp:
        tp.add(data_file)

    tf_arc2 = base_path / "tf_arc2.tar.xz"
    with TarFile.open(tf_arc2, "w:xz") as tp:
        tp.add(data_file)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_single_file_source_date_epoch(base_path, monkeypatch):
    """Writing the same file with different mtime with SOURCE_DATE_EPOCH set produces the
    same hash."""

    data_file = file_factory(base_path)

    arc_base = base_path / "base.tar"
    with ReproducibleTarFile.open(arc_base, "w") as tp:
        tp.add(data_file)

    # With SOURCE_DATE_EPOCH set
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "1691732367")
    arc_sde1 = base_path / "with_sde1.tar"
    with ReproducibleTarFile.open(arc_sde1, "w") as tp:
        tp.add(data_file)

    sleep(2)
    data_file.touch()

    arc_sde2 = base_path / "with_sde2.tar"
    with ReproducibleTarFile.open(arc_sde2, "w") as tp:
        tp.add(data_file)

    # All four archives should have identical content
    assert_archive_contents_equals(arc_base, arc_sde1)
    assert_archive_contents_equals(arc_base, arc_sde2)

    # Base archive hash should match neither, two archives with SOURCE_DATE_EPOCH should match
    assert hash_file(arc_base) != hash_file(arc_sde1)
    assert hash_file(arc_sde1) == hash_file(arc_sde2)


def test_add_single_file_file_mode_env_var(rel_path, monkeypatch):
    """REPRO_TARFILE_FILE_MODE environment variable works."""

    with umask(0o002):
        # Expect 664
        data_file = file_factory(rel_path)

    monkeypatch.setenv("REPRO_TARFILE_FILE_MODE", "600")  # rw-------

    arc_path = rel_path / "archive.tar"
    with ReproducibleTarFile.open(arc_path, "w") as tp:
        tp.add(data_file)

    with TarFile.open(arc_path, "r") as tp:
        print(tp.getmembers())
        mode = tp.getmember(data_file.name).mode & 0o777

    assert mode == 0o600, (oct(mode), oct(0o600))


def test_add_single_dir_dir_mode_env_var(rel_path, monkeypatch):
    """REPRO_TARFILE_DIR_MODE environment variable works."""

    with umask(0o002):
        # Expect 775
        dir_path = rel_path / data_factory()
        dir_path.mkdir()

    monkeypatch.setenv("REPRO_TARFILE_DIR_MODE", "700")  # rwx------

    arc_path = rel_path / "archive.tar"
    with ReproducibleTarFile.open(arc_path, "w") as tp:
        tp.add(dir_path)

    with TarFile.open(arc_path, "r") as tp:
        print(tp.getmembers())
        if sys.version_info >= (3, 9):
            expected_name = dir_path.name + "/"
        else:
            expected_name = dir_path.name
        mode = tp.getmember(expected_name).mode & 0o777

    assert mode == 0o700, (oct(mode), oct(0o700))


def test_add_single_file_uid_env_var(rel_path, monkeypatch):
    """REPRO_TARFILE_UID environment variable works."""

    data_file = file_factory(rel_path)

    monkeypatch.setenv("REPRO_TARFILE_UID", "9999")

    arc_path = rel_path / "archive.tar"
    with ReproducibleTarFile.open(arc_path, "w") as tp:
        tp.add(data_file)

    with TarFile.open(arc_path, "r") as tp:
        print(tp.getmembers())
        uid = tp.getmember(data_file.name).uid

    assert uid == 9999


def test_add_single_file_gid_env_var(rel_path, monkeypatch):
    """REPRO_TARFILE_GID environment variable works."""

    data_file = file_factory(rel_path)

    monkeypatch.setenv("REPRO_TARFILE_GID", "9999")

    arc_path = rel_path / "archive.tar"
    with ReproducibleTarFile.open(arc_path, "w") as tp:
        tp.add(data_file)

    with TarFile.open(arc_path, "r") as tp:
        print(tp.getmembers())
        gid = tp.getmember(data_file.name).gid

    assert gid == 9999


def test_add_single_file_uname_env_var(rel_path, monkeypatch):
    """REPRO_TARFILE_UNAME environment variable works."""

    data_file = file_factory(rel_path)

    monkeypatch.setenv("REPRO_TARFILE_UNAME", "testuser123")

    arc_path = rel_path / "archive.tar"
    with ReproducibleTarFile.open(arc_path, "w") as tp:
        tp.add(data_file)

    with TarFile.open(arc_path, "r") as tp:
        print(tp.getmembers())
        uname = tp.getmember(data_file.name).uname

    assert uname == "testuser123"


def test_add_single_file_gname_env_var(rel_path, monkeypatch):
    """REPRO_TARFILE_GNAME environment variable works."""

    data_file = file_factory(rel_path)

    monkeypatch.setenv("REPRO_TARFILE_GNAME", "testgroup123")

    arc_path = rel_path / "archive.tar"
    with ReproducibleTarFile.open(arc_path, "w") as tp:
        tp.add(data_file)

    with TarFile.open(arc_path, "r") as tp:
        print(tp.getmembers())
        gname = tp.getmember(data_file.name).gname

    assert gname == "testgroup123"


def test_add_single_file_string_paths(rel_path):
    """Writing the same file with different mtime produces the same hash, using string inputs
    instead of Path."""
    data_file = file_factory(rel_path)
    file_name = data_file.name
    assert isinstance(file_name, str)

    rptf_arc1 = rel_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        tp.add(file_name)

    tf_arc1 = rel_path / "tf_arc1.tar"
    with TarFile.open(tf_arc1, "w") as tp:
        tp.add(file_name)

    sleep(2)
    data_file.touch()

    rptf_arc2 = rel_path / "rptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        tp.add(file_name)

    tf_arc2 = rel_path / "tf_arc2.tar"
    with TarFile.open(tf_arc2, "w") as tp:
        tp.add(file_name)

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_add_single_file_arcname(base_path):
    """Writing a single file with explicit arcname."""
    data_file = file_factory(base_path)

    rptf_arc1 = base_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        tp.add(data_file, arcname="lore.txt")

    tf_arc1 = base_path / "tf_arc1.tar"
    with TarFile.open(tf_arc1, "w") as tp:
        tp.add(data_file, arcname="lore.txt")

    sleep(2)
    data_file.touch()

    rptf_arc2 = base_path / "rptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        tp.add(data_file, arcname="lore.txt")

    tf_arc2 = base_path / "tf_arc2.tar"
    with TarFile.open(tf_arc2, "w") as tp:
        tp.add(data_file, arcname="lore.txt")

    # All four archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc1)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)
    assert_archive_contents_equals(rptf_arc1, tf_arc2)

    # ReproducibleTarFile hashes should match; TarFile hashes should not
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)
    assert hash_file(tf_arc1) != hash_file(tf_arc2)


def test_addfile(tmp_path):
    """addfile works as expected"""
    data = data_factory()

    rptf_arc1 = tmp_path / "rptf_arc1.tar"
    with ReproducibleTarFile.open(rptf_arc1, "w") as tp:
        tp.addfile(TarInfo("data.txt"), fileobj=StringIO(data))

    tf_arc = tmp_path / "tf_arc.tar"
    with TarFile.open(tf_arc, "w") as tp:
        tp.addfile(TarInfo("data.txt"), fileobj=StringIO(data))

    sleep(2)

    rptf_arc2 = tmp_path / "rptf_arc2.tar"
    with ReproducibleTarFile.open(rptf_arc2, "w") as tp:
        tp.addfile(TarInfo("data.txt"), fileobj=StringIO(data))

    # All three archives should have identical content
    assert_archive_contents_equals(rptf_arc1, tf_arc)
    assert_archive_contents_equals(rptf_arc1, rptf_arc2)

    # ReproducibleTarFile hashes should match
    assert hash_file(rptf_arc1) == hash_file(rptf_arc2)



from io import BytesIO
import itertools
import logging
from pathlib import Path
import sys
from typing import List, Optional

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

import typer

import repro_tarfile

__version__ = "0.1.0"

app = typer.Typer()


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def version_callback(value: bool):
    if value:
        print(f"repro-tarfile v{repro_tarfile.__version__}")
        print(f"rptar v{__version__}")
        raise typer.Exit()


@app.command(context_settings={"obj": {}})
def rptar(
    in_list: Annotated[List[str], typer.Argument(help="Files to add to the archive.")],
    create: Annotated[bool, typer.Option("--create", "-c", help="Create a new archive.")],
    file: Annotated[
        Optional[str],
        typer.Option(
            "--file",
            "-f",
            help="Path of output archive file.",
        ),
    ] = None,
    gzip: Annotated[bool, typer.Option("--gzip", "-z", help="Use gzip compression.")] = False,
    bzip2: Annotated[bool, typer.Option("--bzip2", "-j", help="Use bzip2 compression.")] = False,
    xz: Annotated[
        bool, typer.Option("--xz", "-J", help="Use xz format with LZMA2 compression.")
    ] = False,
    recursion: Annotated[bool, typer.Option(help="Recurse into directories.")] = True,
    verbose: Annotated[
        int,
        typer.Option(
            "--verbose",
            "-v",
            count=True,
            show_default=False,
            help="Use to increase log verbosity.",
        ),
    ] = 0,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            help="Print version number and exit.",
            callback=version_callback,
        ),
    ] = None,
):
    """A lightweight replacement for `tar -c` for creating tar archives, but reproducibly/
    deterministicly. It supports a subset of common options matching tar.

    Example commands:

    \b
      rptar -czvf archive.tar.gz some_file.txt        # Archive one file
      rptar -czvf archive.tar.gz file1.txt file2.txt  # Archive two files
      rptar -czvf archive.tar.gz some_dir/*.txt       # Archive many files with glob
      rptar -czvf archive.tar.gz some_dir/            # Archive directory recursively
    """
    # Set up logger
    log_level = logging.WARNING - 10 * verbose
    logger.setLevel(log_level)
    log_handler = logging.StreamHandler()
    logger.addHandler(log_handler)
    prog_name = Path(sys.argv[0]).stem
    log_formatter = logging.Formatter(f"%(asctime)s | {prog_name} | %(levelname)s | %(message)s")
    log_handler.setFormatter(log_formatter)

    logger.debug("in_list: %s", in_list)
    logger.debug("create: %s", create)
    logger.debug("file: %s", file)
    logger.debug("gzip: %s", gzip)
    logger.debug("bzip2: %s", bzip2)
    logger.debug("xz: %s", xz)
    logger.debug("recursion: %s", recursion)

    # Check create option
    if not create:
        logger.error("Only create option is supported. Use `tar` for other operations.")
        raise typer.Exit(code=1)

    # Compression
    if sum((gzip, bzip2, xz)) > 1:
        logger.error("Only one compression option can be used at a time.")
        raise typer.Exit(code=1)
    try:
        compression = next(itertools.compress(("gz", "bz2", "xz"), (gzip, bzip2, xz)))
        logger.debug("using compression: %s", compression)
        write_mode = "w:" + compression
    except StopIteration:
        write_mode = "w"

    # Process inputs, manually recurse for logging
    in_paths = set(Path(p) for p in in_list)
    if recursion:
        for path in frozenset(in_paths):
            if path.is_dir():
                in_paths.update(path.glob("**/*"))

    if file:
        out = Path(file).resolve()
        logger.debug("writing to: %s", out)
        with repro_tarfile.open(out, write_mode) as tar:
            for path in sorted(in_paths):
                logger.info("adding: %s", path)
                tar.add(path, recursive=False)
    else:
        with BytesIO() as stream:
            with repro_tarfile.open(fileobj=stream, mode=write_mode) as tar:
                for path in sorted(in_paths):
                    logger.info("adding: %s", path)
                    tar.add(path, recursive=False)
            sys.stdout.buffer.write(stream.getvalue())


if __name__ == "__main__":
    app(prog_name="python -m rptar")



import platform
import subprocess
import sys

import pytest
from typer.testing import CliRunner

from repro_tarfile import __version__ as repro_tarfile_version
from rptar import __version__ as rptar_version
from rptar import app
from tests.utils import assert_archive_contents_equals, dir_tree_factory, file_factory

runner = CliRunner()


@pytest.fixture(autouse=True)
def copyfile_disable(monkeypatch):
    """Disable Apple's ._ file creation."""
    monkeypatch.setenv("COPYFILE_DISABLE", "1")


def test_tar_single_file(base_path):
    data_file = file_factory(base_path)

    rptar_out = base_path / "rptar.tar"
    rptar_args = ["-cf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    tar_out = base_path / "tar.tar"
    tar_cmd = ["tar", "-cf", str(tar_out), str(data_file)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


def test_tar_single_file_gzip(base_path):
    data_file = file_factory(base_path)

    rptar_out = base_path / "rptar.tar.gz"
    rptar_args = ["-czf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    tar_out = base_path / "tar.tar.gz"
    tar_cmd = ["tar", "-czf", str(tar_out), str(data_file)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


def test_tar_single_file_bz2(base_path):
    data_file = file_factory(base_path)

    rptar_out = base_path / "rptar.tar.bz2"
    rptar_args = ["-cjf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    tar_out = base_path / "tar.tar.bz2"
    tar_cmd = ["tar", "-cjf", str(tar_out), str(data_file)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


@pytest.mark.skipif(platform.system() == "Windows", reason="xz not available on Windows")
def test_tar_single_file_xz(base_path):
    data_file = file_factory(base_path)

    rptar_out = base_path / "rptar.tar.xz"
    rptar_args = ["-cJf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    tar_out = base_path / "tar.tar.xz"
    tar_cmd = ["tar", "-cJf", str(tar_out), str(data_file)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


def test_tar_single_file_stdout(base_path):
    data_file = file_factory(base_path)

    rptar_args = ["-c", str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    rptar_out = base_path / "rptar.tar"
    with rptar_out.open("wb") as fp:
        fp.write(rptar_result.stdout_bytes)

    tar_out = base_path / "tar.tar"
    tar_cmd = ["tar", "-cf", str(tar_out), str(data_file)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


def test_tar_directory(base_path):
    """Single directory, not recursive."""
    dir_tree = dir_tree_factory(base_path)

    rptar_out = base_path / "rptar.tar"
    rptar_args = ["-cf", str(rptar_out), str(dir_tree)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    tar_out = base_path / "tar.tar"
    tar_cmd = ["tar", "-cf", str(tar_out), str(dir_tree)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


def test_tar_directory_not_recursive(base_path):
    """Single input directory with recursive --no-recursion flag."""
    dir_tree = dir_tree_factory(base_path)

    rptar_out = base_path / "rptar.tar"
    rptar_args = ["-cf", str(rptar_out), "--no-recursion", str(dir_tree)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args

    tar_out = base_path / "tar.tar"
    tar_cmd = ["tar", "-cf", str(tar_out), "--no-recursion", str(dir_tree)]
    tar_result = subprocess.run(tar_cmd)
    assert tar_result.returncode == 0, tar_cmd

    assert_archive_contents_equals(rptar_out, tar_out)


def test_verbosity(rel_path):
    """Adjustment of verbosity with -v and -q."""
    data_file = file_factory(rel_path)
    rptar_out = rel_path / "rptar.tar"

    # Base case, should be WARNING level
    rptar_args = ["-cf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args
    assert "INFO" not in rptar_result.output
    assert "DEBUG" not in rptar_result.output

    # With -v, should be INFO level
    rptar_args = ["-cvf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args
    assert "INFO" in rptar_result.output
    assert "DEBUG" not in rptar_result.output

    # With -vv, should be DEBUG level
    rptar_args = ["-cvvf", str(rptar_out), str(data_file)]
    rptar_result = runner.invoke(app, rptar_args)
    assert rptar_result.exit_code == 0, rptar_args
    assert "INFO" in rptar_result.output
    assert "DEBUG" in rptar_result.output


def test_version():
    """With --version flag."""
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    output_lines = result.output.split("\n")
    assert output_lines[0].startswith("repro-tarfile ")
    assert output_lines[0].endswith(f"v{repro_tarfile_version}")
    assert output_lines[1].startswith("rptar ")
    assert output_lines[1].endswith(f"v{rptar_version}")


def test_python_dash_m_invocation():
    result = subprocess.run(
        [sys.executable, "-m", "rptar", "--help"],
        capture_output=True,
        text=True,
        universal_newlines=True,
    )
    assert result.returncode == 0
    assert "Usage: python -m rptar" in result.stdout



import builtins
import contextlib
import copy
import datetime
import os
from tarfile import CompressionError, ReadError, TarFile, TarInfo
from typing import IO, Optional

__version__ = "0.1.0"

__all__ = [
    "open",
    "ReproducibleTarFile",
    "TarInfo",
]


def mtime() -> int:
    """Returns mtime value used to force overwrite on all TarInfo objects. Defaults to
    315532800 (corresponding to 1980-01-01 00:00:00 UTC). You can set this with the environment
    variable SOURCE_DATE_EPOCH as an integer value representing seconds since Epoch.
    """
    source_date_epoch = os.environ.get("SOURCE_DATE_EPOCH", None)
    if source_date_epoch is not None:
        return int(source_date_epoch)
    return int(datetime.datetime(1980, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc).timestamp())


def file_mode() -> int:
    """Returns the file permissions mode value used to force overwrite on all TarInfo objects.
    Defaults to 0o644 (rw-r--r--). You can set this with the environment variable
    REPRO_TARFILE_FILE_MODE. It should be in the Unix standard three-digit octal representation
    (e.g., '644').
    """
    file_mode_env = os.environ.get("REPRO_TARFILE_FILE_MODE", None)
    if file_mode_env is not None:
        return int(file_mode_env, 8)
    return 0o644


def dir_mode() -> int:
    """Returns the directory permissions mode value used to force overwrite on all TarInfo objects.
    Defaults to 0o755 (rwxr-xr-x). You can set this with the environment variable
    REPRO_TARFILE_DIR_MODE. It should be in the Unix standard three-digit octal representation
    (e.g., '755').
    """
    dir_mode_env = os.environ.get("REPRO_TARFILE_DIR_MODE", None)
    if dir_mode_env is not None:
        return int(dir_mode_env, 8)
    return 0o755


def uid() -> int:
    """Returns the user id value used to force overwrite on all TarInfo objects. Defaults to 0.
    You can set this with the environment variable REPRO_TARFILE_UID.
    """
    uid_env = os.environ.get("REPRO_TARFILE_UID", None)
    if uid_env is not None:
        return int(uid_env)
    return 0


def gid() -> int:
    """Returns the group id value used to force overwrite on all TarInfo objects. Defaults to 0.
    You can set this with the environment variable REPRO_TARFILE_GID.
    """
    gid_env = os.environ.get("REPRO_TARFILE_GID", None)
    if gid_env is not None:
        return int(gid_env)
    return 0


def uname() -> str:
    """Returns the user name value used to force overwrite on all TarInfo objects. Defaults to
    empty string. You can set this with the environment variable REPRO_TARFILE_UNAME.
    """
    uname_env = os.environ.get("REPRO_TARFILE_UNAME", None)
    if uname_env is not None:
        return uname_env
    return ""


def gname() -> str:
    """Returns the group name value used to force overwrite on all TarInfo objects. Defaults to
    empty string. You can set this with the environment variable REPRO_TARFILE_GNAME.
    """
    gname_env = os.environ.get("REPRO_TARFILE_GNAME", None)
    if gname_env is not None:
        return gname_env
    return ""


_NO_TARFILE_ATTR = object()


@contextlib.contextmanager
def _temporarily_delete_tarfile_attr(tarinfo: TarInfo):
    """Context manager for temporarily deleting the tarfile attribute from a TarInfo object.

    For some reason, TarInfo objects returned by TarFile.gettarinfo get assigned a 'tarfile'
    attribute with a reference ot the TarInfo object.

    https://github.com/python/cpython/blob/f7c05d7ad3075a1dbeed86b6b12903032e4afba6/Lib/tarfile.py#L2033

    The TarInfo can have IO streams that prevent deepcopy. This context manager temporarily deletes
    the tarfile attribute so that we can do a deepcopy successfully.
    """
    tarfile_attr = getattr(tarinfo, "tarfile", _NO_TARFILE_ATTR)
    if tarfile_attr is not _NO_TARFILE_ATTR:
        delattr(tarinfo, "tarfile")
    try:
        yield
    finally:
        if tarfile_attr is not _NO_TARFILE_ATTR:
            # mypy doesn't handle seninel objects
            # https://github.com/python/mypy/issues/15788
            tarinfo.tarfile = tarfile_attr  # type: ignore[assignment]


class ReproducibleTarFile(TarFile):
    # Following method modified from Python 3.12
    # https://github.com/python/cpython/blob/09b8b14e05557304ab12870137181685c2dcbe25/Lib/tarfile.py#L1856-L1887
    # Copyright Python Software Foundation, licensed under PSF License Version 2
    # See LICENSE file for full license agreement and notice of copyright
    @classmethod
    def gzopen(cls, name, mode="r", fileobj=None, compresslevel=9, **kwargs):
        """Open gzip compressed tar archive name for reading or writing.
        Appending is not allowed.
        """
        if mode not in ("r", "w", "x"):
            raise ValueError("mode must be 'r', 'w' or 'x'")

        try:
            from gzip import GzipFile
        except ImportError:
            raise CompressionError("gzip module is not available") from None

        try:
            ## repro-tarfile MODIFIED ##
            # Overwrite filename and mtime when initializing GzipFile
            if fileobj is None:
                fileobj = builtins.open(name, mode + "b")
            fileobj = GzipFile("", mode + "b", compresslevel, fileobj, mtime=mtime())
            #########################
        except OSError as e:
            if fileobj is not None and mode == "r":
                raise ReadError("not a gzip file") from e
            raise

        try:
            t = cls.taropen(name, mode, fileobj, **kwargs)
        except OSError as e:
            fileobj.close()
            if mode == "r":
                raise ReadError("not a gzip file") from e
            raise
        except:
            fileobj.close()
            raise
        t._extfileobj = False
        return t

    def addfile(self, tarinfo: TarInfo, fileobj: Optional[IO[bytes]] = None) -> None:
        """Add the TarInfo object `tarinfo' to the archive. If `fileobj' is
        given, it should be a binary file, and tarinfo.size bytes are read
        from it and added to the archive. You can create TarInfo objects
        directly, or by using gettarinfo().
        """
        if tarinfo.isdir():
            mode = 0o40000 | dir_mode()
        else:
            mode = 0o100000 | file_mode()
        # See docstring for _temporarily_delete_tarfile_attr for why we need to do this.
        with _temporarily_delete_tarfile_attr(tarinfo):
            try:
                tarinfo_copy = tarinfo.replace(
                    mtime=mtime(),
                    mode=mode,
                    uid=uid(),
                    gid=gid(),
                    uname=uname(),
                    gname=gname(),
                    deep=True,
                )
            except AttributeError as e:
                # Some older versions of Python don't have replace method
                # Added in: 3.8.17, 3.9.17, 3.10.12, 3.11.4, 3.12
                if "'TarInfo' object has no attribute 'replace'" in str(e):
                    tarinfo_copy = copy.deepcopy(tarinfo)
                    tarinfo_copy.mtime = mtime()
                    tarinfo_copy.mode = mode
                    tarinfo_copy.uid = uid()
                    tarinfo_copy.gid = gid()
                    tarinfo_copy.uname = uname()
                    tarinfo_copy.gname = gname()
                else:
                    raise
        return super().addfile(tarinfo=tarinfo_copy, fileobj=fileobj)


open = ReproducibleTarFile.open



"""Basic usage example of repro_tarfile.
"""

from io import StringIO
from pathlib import Path

from repro_tarfile import ReproducibleTarFile, TarInfo

example_dir = Path(__file__).parent

with ReproducibleTarFile.open(example_dir / "archive.tar.gz", "w:gz") as tar:
    # Use add to add a file to the archive
    tar.add(example_dir / "data.txt", arcname="data.txt")
    # Or addfile to write from a data stream to the archive
    tar.addfile(TarInfo(name="lore.txt"), fileobj=StringIO("goodbye"))
